(* D'aprs le code Lucid Synchrone de Gwenael Delaval *)

(*********************** DECODEUR GSM AVEC BUFFERS ****************************)
(******************************************************************************)

open Gsm_encoding

(********** cyclic decoding **********)

(* stutter50 *)
let node stutter x = 
  merge (10) x (buffer(x))

let node stutter16 x =
  stutter (stutter (stutter (stutter x)))

let node stutter32 x =
  stutter (stutter (stutter (stutter (stutter x))))

let node stutter48 x =
  merge (1^32 0^16) (stutter32 x) (buffer(stutter16 x))

let node stutter49 x =
 merge (1^48 0) (stutter48 x) (buffer(x))

let node stutter50 x =
  merge (1^49 0) (stutter49 x) (buffer(x))

let node split3 x = x when (100), x when (010), x when (001)

let node and_by_3 x = y where
  rec x1, x2, x3 = split3 x
  and y = ((buffer ((buffer x1) && x2)) && x3)

let node cyclic_decoding x = o where
  rec (data, redondancy) = 
    (x when (1^52 0) when (1^51 0) when (1^50 0) , x whenot (1^50 0^3))
  and reencoded_data = cyclic_encoding data
  and code = reencoded_data whenot (1^50 0^3)
  and ok = (code = redondancy) 
  and frame_ok = and_by_3 ok
  and data_ok = (stutter50 (buffer(frame_ok)))
  and current_frame = buffer(data)
  and last_frame = merge 1^50(0) true (buffer(current_frame))
  and o = if data_ok then current_frame else last_frame

(********** convolutional decoding **********) 
(* hamming distance *)
let node hamming_dist ((x1,x2),(y1,y2)) =
  (if x1 = y1 then 0 else 1) + (if x2 = y2 then 0 else 1)

let node convolutional_decoding x = (* non implemente *)
  (x when (01)) when 0^188(1^185 0^4) 
(* code factice, avec le bon type de sortie *)

(********** join speech **********) 
let node join_speech (x_Ia, x_Ib, x_II) = x where
  rec x = merge (1^50 0^210) x_Ia (merge (1^132 0^78) x_Ib x_II)

(********** split encoded speech **********) 
let node split_I_II x = (x when (1^378 0^78), x whenot (1^378 0^78))
let node split_Ia_Ib x_I = (x_I when (1^53 0^132), x_I whenot (1^53 0^132))

(********** GSM speech decoding **********)

(* les buffers sont indispensables, 
   la version sans buffers est mal type dans l'abstrait *)

(* version retenue *)
(* l'ajustement des prfixes choue dans la resolution concrete *)
let node decode_with_3_buffers x = y where 
  rec x_I_enc, x_II = split_I_II x
  and x_I = convolutional_decoding x_I_enc
  and x_Ia_enc, x_Ib = split_Ia_Ib x_I
  and x_Ia = cyclic_decoding x_Ia_enc
  and y = join_speech (buffer(x_Ia), buffer(x_Ib), buffer(x_II))


(********** identity **********)
let node id_speech x = y where
  rec y = decode_with_3_buffers(encode_with_3_buffers(x))
